@@ -1,5 +1,6 @@
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE CPP #-}
 module Codec.Picture.Jpg.Common
     ( DctCoefficients
     , JpgUnpackerParameter( .. )
@@ -18,7 +19,10 @@ module Codec.Picture.Jpg.Common
     , decodeRestartInterval
     ) where
 
-import Control.Applicative( (<$>), pure )
+#if !MIN_VERSION_base(4,8,0)
+import Control.Applicative( pure, (<$>) )
+#endif
+
 import Control.Monad( replicateM, when )
 import Control.Monad.ST( ST, runST )
 import Data.Bits( unsafeShiftL, unsafeShiftR, (.&.) )
